@@ -25,9 +25,10 @@
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
-#include <stdlib.h>
+#include <inttypes.h>
 #include <signal.h>
+
+#include <unistd.h>
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <sys/un.h>
@@ -51,21 +52,29 @@
 #include "error.h"
 #include "encoding.h"
 #include <libvchan.h>
+#include <poll.h>
+
+
+#include <linux/input.h>
+#include <linux/uinput.h>
+#include <time.h>
+
+
 
 /* Time in milliseconds after which the clipboard data should be wiped */
 #define CLIPBOARD_WIPE_TIME 60000
 
 /* Get the size of an array.  Error out on pointers. */
 #define QUBES_ARRAY_SIZE(x) (0 * sizeof(struct { \
-    int tried_to_compute_number_of_array_elements_in_a_pointer: \
-        1 - 2*__builtin_types_compatible_p(__typeof__(x), __typeof__(&((x)[0]))); \
+    uint8_t tried_to_compute_number_of_array_elements_in_a_pointer: \
+        8 - 16*__builtin_types_compatible_p(__typeof__(x), __typeof__(&((x)[0]))); \
     }) + sizeof(x)/sizeof((x)[0]))
 #define SOCKET_ADDRESS  "/var/run/xf86-qubes-socket"
 
 /* Supported protocol version */
 
 #define PROTOCOL_VERSION_MAJOR 1
-#define PROTOCOL_VERSION_MINOR 3
+#define PROTOCOL_VERSION_MINOR 7
 #define PROTOCOL_VERSION (PROTOCOL_VERSION_MAJOR << 16 | PROTOCOL_VERSION_MINOR)
 
 #if !(PROTOCOL_VERSION_MAJOR == QUBES_GUID_PROTOCOL_VERSION_MAJOR && \
@@ -79,14 +88,14 @@
 #  define UNUSED(x) UNUSED_ ## x
 #endif
 
-int damage_event, damage_error;
-int xfixes_event, xfixes_error;
+static int damage_event, damage_error;
+static int xfixes_event, xfixes_error;
 /* from gui-common/error.c */
 extern int print_x11_errors;
 
-char **saved_argv;
+static char **saved_argv;
 
-struct _global_handles {
+typedef struct {
     Display *display;
     int screen;            /* shortcut to the default screen */
     Window root_win;       /* root attributes */
@@ -124,8 +133,12 @@ struct _global_handles {
     int composite_redirect_automatic;
     pid_t x_pid;
     uint32_t domid;
+    uint32_t protocol_version;
     Time time;
-};
+    int uinput_fd;
+    int created_input_device;
+    uint8_t last_known_modifier_states;
+} Ghandles;
 
 struct window_data {
     int is_docked; /* is it docked icon window */
@@ -140,15 +153,14 @@ struct embeder_data {
     XID icon_window;
 };
 
-struct genlist *windows_list;
-struct genlist *embeder_list;
-typedef struct _global_handles Ghandles;
-Ghandles *ghandles_for_vchan_reinitialize;
+static struct genlist *windows_list;
+static struct genlist *embeder_list;
+static Ghandles *ghandles_for_vchan_reinitialize;
 
 #define SKIP_NONMANAGED_WINDOW do {                                    \
     if (!list_lookup(windows_list, window)) {                          \
         if (g->log_level > 0)                                          \
-            fprintf(stderr, "Skipping unmanaged window 0x%x",          \
+            fprintf(stderr, "Skipping unmanaged window 0x%x\n",        \
                     (int) window);                                     \
         return;                                                        \
     }                                                                  \
@@ -277,12 +289,34 @@ static struct supported_cursor supported_cursors[] = {
 
 #define NUM_SUPPORTED_CURSORS (QUBES_ARRAY_SIZE(supported_cursors))
 
-static int compare_supported_cursors(const void *a,
-                              const void *b) {
-    return strcmp(((struct supported_cursor *)a)->name,
-                  ((struct supported_cursor *)b)->name);
+static int compare_supported_cursors(const void *a, const void *b) {
+    return strcmp(((const struct supported_cursor *)a)->name,
+                  ((const struct supported_cursor *)b)->name);
 }
 
+
+static void send_event(Ghandles * g, const struct input_event *iev) {
+    int status = write(g->uinput_fd, iev, sizeof(struct input_event));
+    if ( status < 0 ) {
+        if (g->log_level > 0) {
+            fprintf(stderr, "write failed, falling back to xdriver. TYPE:%d CODE:%d VALUE: %d WRITE ERROR CODE: %d\n", iev->type, iev->code, iev->value, status);
+        }
+        g->created_input_device = 0;
+    }
+    
+    // send syn
+    const struct input_event syn = {.type = EV_SYN, .code = 0, .value = 0};
+    status = write(g->uinput_fd, &(syn), sizeof(struct input_event));
+    
+    if ( status < 0 ) {
+        if (g->log_level > 0) {
+            fprintf(stderr, "writing SYN failed, falling back to xdriver. WRITE ERROR CODE: %d\n", status);
+        }
+        g->created_input_device = 0;
+    }
+}
+
+
 static void send_wmname(Ghandles * g, XID window);
 static void send_wmnormalhints(Ghandles * g, XID window, int ignore_fail);
 static void send_wmclass(Ghandles * g, XID window, int ignore_fail);
@@ -396,7 +430,7 @@ static void process_xevent_createnotify(Ghandles * g, XCreateWindowEvent * ev)
                 "handle_create, ret=0x%x\n", (int) ev->window,
                 ret);
         return;
-    };
+    }
 
     if (g->log_level > 0)
         fprintf(stderr, "Create for 0x%x class 0x%x\n",
@@ -408,7 +442,7 @@ static void process_xevent_createnotify(Ghandles * g, XCreateWindowEvent * ev)
     }
 
     if (ev->parent != g->root_win) {
-        /* GUI daemon won’t support this much longer */
+        /* GUI daemon no longer supports this */
         fprintf(stderr,
                 "CREATE with non-root parent window 0x%x\n",
                 (unsigned int)ev->parent);
@@ -467,23 +501,19 @@ static void process_xevent_createnotify(Ghandles * g, XCreateWindowEvent * ev)
 static void feed_xdriver(Ghandles * g, int type, int arg1, int arg2)
 {
     char ans;
-    int ret;
+    ssize_t ret;
     struct xdriver_cmd cmd;
 
     cmd.type = type;
     cmd.arg1 = arg1;
     cmd.arg2 = arg2;
-    if (write(g->xserver_fd, &cmd, sizeof(cmd)) != sizeof(cmd)) {
-        perror("unix write");
-        exit(1);
-    }
+    if (write(g->xserver_fd, &cmd, sizeof(cmd)) != sizeof(cmd))
+        err(1, "unix write");
     ans = '1';
     ret = read(g->xserver_fd, &ans, 1);
     if (ret != 1 || ans != '0') {
         perror("unix read");
-        fprintf(stderr, "read returned %d, char read=0x%x\n", ret,
-                (int) ans);
-        exit(1);
+        err(1, "read returned %zd, char read=0x%x\n", ret, (int) ans);
     }
 }
 
@@ -496,16 +526,14 @@ void send_pixmap_grant_refs(Ghandles * g, XID window)
     int ret;
 
     feed_xdriver(g, 'W', (int) window, 0);
-    if (read(g->xserver_fd, &wd_msg_len, sizeof(wd_msg_len)) != sizeof(wd_msg_len)) {
-        perror("unix read wd_msg_len");
-        exit(1);
-    }
+    if (read(g->xserver_fd, &wd_msg_len, sizeof(wd_msg_len)) != sizeof(wd_msg_len))
+        err(1, "unix read wd_msg_len");
     if (wd_msg_len == 0) {
         fprintf(stderr, "Failed to get window dump for window 0x%lx\n",
                 window);
         return;
     }
-    wd_msg_buf = alloca(wd_msg_len);
+    wd_msg_buf = malloc(wd_msg_len);
     if (!wd_msg_buf) {
         fprintf(stderr, "Failed to allocate memory for window dump 0x%lx\n",
                 window);
@@ -514,14 +542,10 @@ void send_pixmap_grant_refs(Ghandles * g, XID window)
     rcvd = 0;
     while (rcvd < wd_msg_len) {
         ret = read(g->xserver_fd, wd_msg_buf + rcvd, wd_msg_len - rcvd);
-        if (ret == 0) {
-            fprintf(stderr, "unix read EOF\n");
-            exit(1);
-        }
-        if (ret < 0) {
-            perror("unix read error");
-            exit(1);
-        }
+        if (ret == 0)
+            errx(1, "unix read EOF");
+        if (ret < 0)
+            err(1, "unix read error");
         rcvd += ret;
     }
     hdr.type = MSG_WINDOW_DUMP;
@@ -529,6 +553,9 @@ void send_pixmap_grant_refs(Ghandles * g, XID window)
     hdr.untrusted_len = wd_msg_len;
     write_struct(g->vchan, hdr);
     write_data(g->vchan, (char *) wd_msg_buf, wd_msg_len);
+    free(wd_msg_buf);
+    if (g->protocol_version < QUBES_GUID_MIN_MSG_WINDOW_DUMP_ACK)
+        feed_xdriver(g, 'a', 0, 0);
 }
 
 /* return 1 on success, 0 otherwise */
@@ -640,7 +667,7 @@ void retrieve_wmprotocols(Ghandles * g, XID window, int ignore_fail)
     struct genlist *l;
 
     if (!((l=list_lookup(windows_list, window)) && (l->data))) {
-        fprintf(stderr, "ERROR retrieve_wmprotocols: Window 0x%x data not initialized", (int)window);
+        fprintf(stderr, "ERROR retrieve_wmprotocols: Window 0x%x data not initialized\n", (int)window);
         return;
     }
 
@@ -676,7 +703,7 @@ void retrieve_wmhints(Ghandles * g, XID window, int ignore_fail)
     struct genlist *l;
 
     if (!((l=list_lookup(windows_list, window)) && (l->data))) {
-        fprintf(stderr, "ERROR retrieve_wmhints: Window 0x%x data not initialized", (int)window);
+        fprintf(stderr, "ERROR retrieve_wmhints: Window 0x%x data not initialized\n", (int)window);
         return;
     }
 
@@ -931,7 +958,7 @@ static void process_xevent_configure(Ghandles * g, XID window,
                     "XGetWindowAttributes for 0x%x failed in "
                     "handle_xevent_configure, ret=0x%x\n", (int) ((struct window_data*)l->data)->embeder, ret);
             return;
-        };
+        }
         if (ev->x != 0 || ev->y != 0 || ev->width != attr.width || ev->height != attr.height) {
             XMoveResizeWindow(g->display, window, 0, 0, attr.width, attr.height);
         }
@@ -1178,7 +1205,7 @@ static void process_xevent_message(Ghandles * g, XClientMessageEvent * ev)
                 w = ev->data.l[2];
 
                 if (!(l=list_lookup(windows_list, w))) {
-                    fprintf(stderr, "ERROR process_xevent_message: Window 0x%x not initialized", (int)w);
+                    fprintf(stderr, "ERROR process_xevent_message: Window 0x%x not initialized\n", (int)w);
                     return;
                 }
                 if (g->log_level > 0)
@@ -1203,7 +1230,7 @@ static void process_xevent_message(Ghandles * g, XClientMessageEvent * ev)
                     XFree(data);
 
                 if (!(l->data)) {
-                    fprintf(stderr, "ERROR process_xevent_message: Window 0x%x data not initialized", (int)w);
+                    fprintf(stderr, "ERROR process_xevent_message: Window 0x%x data not initialized\n", (int)w);
                     return;
                 }
                 wd = (struct window_data*)(l->data);
@@ -1235,7 +1262,7 @@ static void process_xevent_message(Ghandles * g, XClientMessageEvent * ev)
                             "handle_dock, ret=0x%x\n", (int) w,
                             ret);
                     return;
-                };
+                }
 
                 memset(&resp, 0, sizeof(resp));
                 resp.type = ClientMessage;
@@ -1388,14 +1415,14 @@ static int send_full_window_info(Ghandles *g, XID w, struct window_data *wd)
                 "send_window_state, ret=0x%x\n", (int) w,
                 ret);
         return 0;
-    };
+    }
     ret = XQueryTree(g->display, window_to_query, &root, &parent, &children_list, &children_count);
     if (ret != 1) {
         fprintf(stderr, "XQueryTree for 0x%x failed in "
                 "send_window_state, ret=0x%x\n", (int) w,
                 ret);
         return 0;
-    };
+    }
     if (children_list)
         XFree(children_list);
     if (parent != g->root_win) {
@@ -1460,6 +1487,7 @@ static void send_all_windows_info(Ghandles *g) {
     struct genlist *curr = windows_list->next;
     int ret;
 
+    feed_xdriver(g, 'A', 0, 0);
     while (curr != windows_list) {
         ret = send_full_window_info(g, curr->key, (struct window_data *)curr->data);
         curr = curr->next;
@@ -1620,62 +1648,129 @@ static void handle_keypress(Ghandles * g, XID UNUSED(winid))
     struct msg_keypress key;
     XkbStateRec state;
     read_data(g->vchan, (char *) &key, sizeof(key));
-    // sync modifiers state
-    if (XkbGetState(g->display, XkbUseCoreKbd, &state) != Success) {
-        if (g->log_level > 0)
-            fprintf(stderr, "failed to get modifier state\n");
-        state.mods = key.state;
-    }
-    if (!g->sync_all_modifiers) {
-        // ignore all but CapsLock
-        state.mods &= LockMask;
-        key.state &= LockMask;
-    }
-    if (state.mods != key.state) {
-        XModifierKeymap *modmap;
-        int mod_index;
-        int mod_mask;
 
-        modmap = XGetModifierMapping(g->display);
-        if (!modmap) {
+    if(!g->created_input_device) {
+        // sync modifiers state
+        if (XkbGetState(g->display, XkbUseCoreKbd, &state) != Success) {
             if (g->log_level > 0)
-                fprintf(stderr, "failed to get modifier mapping\n");
+                fprintf(stderr, "failed to get modifier state\n");
+            state.mods = key.state;
+        }
+        if (!g->sync_all_modifiers) {
+            // ignore all but CapsLock
+            state.mods &= LockMask;
+            key.state &= LockMask;
+        }
+        if (state.mods != key.state) {
+            XModifierKeymap *modmap;
+            int mod_index;
+            int mod_mask;
+
+            modmap = XGetModifierMapping(g->display);
+            if (!modmap) {
+                if (g->log_level > 0)
+                    fprintf(stderr, "failed to get modifier mapping\n");
+            } else {
+                // from X.h:
+                // #define ShiftMapIndex           0
+                // #define LockMapIndex            1
+                // #define ControlMapIndex         2
+                // #define Mod1MapIndex            3
+                // #define Mod2MapIndex            4
+                // #define Mod3MapIndex            5
+                // #define Mod4MapIndex            6
+                // #define Mod5MapIndex            7
+                for (mod_index = 0; mod_index < 8; mod_index++) {
+                    if (modmap->modifiermap[mod_index*modmap->max_keypermod] == 0x00) {
+                        if (g->log_level > 1)
+                            fprintf(stderr, "ignoring disabled modifier %d\n", mod_index);
+                        // no key set for this modifier, ignore
+                        continue;
+                    }
+                    mod_mask = (1<<mod_index);
+                    // special case for caps lock switch by press+release
+                    if (mod_index == LockMapIndex) {
+                        if ((state.mods & mod_mask) ^ (key.state & mod_mask)) {
+                            feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 1);
+                            feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 0);
+                        }
+                    } else {
+                        if ((state.mods & mod_mask) && !(key.state & mod_mask))
+                            feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 0);
+                        else if (!(state.mods & mod_mask) && (key.state & mod_mask))
+                            feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 1);
+                    }
+                }
+                XFreeModifiermap(modmap);
+            }
+        }
+        
+        feed_xdriver(g, 'K', key.keycode, key.type == KeyPress ? 1 : 0);
+    } else {
+        int mod_mask;
+        int mod_index;
+        struct input_event iev;
+        iev.type = EV_KEY;
+        XModifierKeymap *modmap;
+        modmap = XGetModifierMapping(g->display);
+        
+        if (!modmap) {
+                if (g->log_level > 0)
+                    fprintf(stderr, "failed to get modifier mapping\n");
         } else {
-            // from X.h:
-            // #define ShiftMapIndex           0
-            // #define LockMapIndex            1
-            // #define ControlMapIndex         2
-            // #define Mod1MapIndex            3
-            // #define Mod2MapIndex            4
-            // #define Mod3MapIndex            5
-            // #define Mod4MapIndex            6
-            // #define Mod5MapIndex            7
-            for (mod_index = 0; mod_index < 8; mod_index++) {
+            for(mod_index = 0; mod_index < 8; mod_index++) {
                 if (modmap->modifiermap[mod_index*modmap->max_keypermod] == 0x00) {
-                    if (g->log_level > 1)
-                        fprintf(stderr, "ignoring disabled modifier %d\n", mod_index);
-                    // no key set for this modifier, ignore
-                    continue;
+                        if (g->log_level > 1)
+                            fprintf(stderr, "ignoring disabled modifier %d\n", mod_index);
+                        // no key set for this modifier, ignore
+                        continue;
                 }
                 mod_mask = (1<<mod_index);
                 // special case for caps lock switch by press+release
                 if (mod_index == LockMapIndex) {
-                    if ((state.mods & mod_mask) ^ (key.state & mod_mask)) {
-                        feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 1);
-                        feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 0);
+                    if ((g->last_known_modifier_states & mod_mask) ^ (key.state & mod_mask)) {
+                        iev.code = modmap->modifiermap[mod_index*modmap->max_keypermod] - 8;
+                        iev.value = 1;
+                        send_event(g, &iev);
+                        iev.value = 0;
+                        send_event(g, &iev);
+                        // update state for caps_lock
+                        g->last_known_modifier_states ^= mod_mask;
                     }
                 } else {
-                    if ((state.mods & mod_mask) && !(key.state & mod_mask))
-                        feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 0);
-                    else if (!(state.mods & mod_mask) && (key.state & mod_mask))
-                        feed_xdriver(g, 'K', modmap->modifiermap[mod_index*modmap->max_keypermod], 1);
+                    // last modifier state was pressed down, modifier has since been released
+                    if ((g->last_known_modifier_states & mod_mask) && !(key.state & mod_mask)) {
+                        iev.code = modmap->modifiermap[mod_index*modmap->max_keypermod] - 8;
+                        iev.value = 0;
+                        // send modifier release
+                        send_event(g, &iev);
+                        // update state for this modifier
+                        g->last_known_modifier_states ^= mod_mask;
+                    }
+                    
+                    // last modifier state was up, modifier has since been pressed down
+                    else if (!(g->last_known_modifier_states & mod_mask) && (key.state & mod_mask)) {
+                        iev.code = modmap->modifiermap[mod_index*modmap->max_keypermod] - 8;
+                        iev.value = 1;
+                        // send modifier press
+                        send_event(g, &iev);
+                        // update state for this modifier
+                        g->last_known_modifier_states ^= mod_mask;
+                    }
                 }
             }
-            XFreeModifiermap(modmap);
         }
-    }
+        
+        XFreeModifiermap(modmap);
 
-    feed_xdriver(g, 'K', key.keycode, key.type == KeyPress ? 1 : 0);
+        // caps lock needs to be excluded to not send down, up, down or down, up, up on a caps lock sync instead of down, up
+        if(key.keycode-8 != KEY_CAPSLOCK) {
+            iev.code = key.keycode-8;
+            iev.value = (key.type == KeyPress ? 1 : 0);
+            send_event(g, &iev);
+        }
+        
+    }
 }
 
 static void handle_button(Ghandles * g, XID winid)
@@ -1717,7 +1812,7 @@ static void handle_motion(Ghandles * g, XID winid)
                 "XGetWindowAttributes for 0x%x failed in "
                 "do_button, ret=0x%x\n", (int) winid, ret);
         return;
-    };
+    }
 
     feed_xdriver(g, 'M', attr.x + key.x, attr.y + key.y);
 }
@@ -1742,15 +1837,16 @@ static void handle_crossing(Ghandles * g, XID winid)
 
     if (key.mode != NotifyNormal)
         return;
-    ret = XGetWindowAttributes(g->display, winid, &attr);
-    if (ret != 1) {
-        fprintf(stderr,
-                "XGetWindowAttributes for 0x%x failed in "
-                "handle_crossing, ret=0x%x\n", (int) winid, ret);
-        return;
-    };
 
     if (key.type == EnterNotify) {
+        ret = XGetWindowAttributes(g->display, winid, &attr);
+        if (ret != 1) {
+            fprintf(stderr,
+                    "XGetWindowAttributes for 0x%x failed in "
+                    "handle_crossing, ret=0x%x\n", (int) winid, ret);
+            return;
+        }
+
         // hide stub window
         XUnmapWindow(g->display, g->stub_win);
         feed_xdriver(g, 'M', attr.x + key.x, attr.y + key.y);
@@ -1821,7 +1917,7 @@ static void handle_focus(Ghandles * g, XID winid)
             if (((struct window_data*)l->data)->is_docked)
                 XRaiseWindow(g->display, ((struct window_data*)l->data)->embeder);
         } else {
-            fprintf(stderr, "WARNING handle_focus: Window 0x%x data not initialized", (int)winid);
+            fprintf(stderr, "WARNING handle_focus: FocusIn: Window 0x%x data not initialized\n", (int)winid);
             input_hint = True;
             use_take_focus = False;
         }
@@ -1842,7 +1938,7 @@ static void handle_focus(Ghandles * g, XID winid)
         if ( (l=list_lookup(windows_list, winid)) && (l->data) )
             input_hint = ((struct window_data*)l->data)->input_hint;
         else {
-            fprintf(stderr, "WARNING handle_focus: Window 0x%x data not initialized", (int)winid);
+            fprintf(stderr, "WARNING handle_focus: FocusOut: Window 0x%x data not initialized\n", (int)winid);
             input_hint = True;
         }
         if (input_hint)
@@ -1919,7 +2015,7 @@ static void handle_close(Ghandles * g, XID winid)
     if ( (l=list_lookup(windows_list, winid)) && (l->data) ) {
         use_delete_window = ((struct window_data*)l->data)->support_delete_window;
     } else {
-        fprintf(stderr, "WARNING handle_close: Window 0x%x data not initialized",
+        fprintf(stderr, "WARNING handle_close: Window 0x%x data not initialized\n",
                 (int)winid);
         use_delete_window = True; /* gentler, though it may be a no-op */
     }
@@ -2125,6 +2221,15 @@ static void handle_message(Ghandles * g)
         case MSG_WINDOW_FLAGS:
             handle_window_flags(g, hdr.window);
             break;
+        case MSG_DESTROY:
+            /* currently not used */
+            break;
+        case MSG_WINDOW_DUMP_ACK:
+            if (g->protocol_version >= QUBES_GUID_MIN_MSG_WINDOW_DUMP_ACK) {
+                feed_xdriver(g, 'a', 0, 0);
+                break;
+            }
+            __attribute__((fallthrough));
         default:
             fprintf(stderr, "got unknown msg type %d, ignoring\n", hdr.type);
             while (hdr.untrusted_len > 0) {
@@ -2153,13 +2258,24 @@ static pid_t get_xconf_and_run_x(Ghandles *g)
     return x_pid;
 }
 
-static void send_protocol_version(libvchan_t *vchan)
+static void handshake(Ghandles *g)
 {
     uint32_t version = PROTOCOL_VERSION;
-    write_struct(vchan, version);
+    write_struct(g->vchan, version);
+    version = 0;
+    read_struct(g->vchan, version);
+    uint16_t major_version = version >> 16, minor_version = version & 0xFFFF;
+    if (version > PROTOCOL_VERSION ||
+        major_version != PROTOCOL_VERSION_MAJOR ||
+        minor_version < 4)
+        errx(2, "Incompatible GUI daemon version (offered %" PRIu16 ".%" PRIu16
+                ", got %" PRIu16 ".%" PRIu16 ")",
+                PROTOCOL_VERSION_MAJOR, PROTOCOL_VERSION_MINOR,
+                major_version, minor_version);
+    g->protocol_version = version;
 }
 
-static void handle_guid_disconnect()
+static void handle_guid_disconnect(void)
 {
     Ghandles *g = ghandles_for_vchan_reinitialize;
     struct msg_xconf xconf;
@@ -2170,25 +2286,25 @@ static void handle_guid_disconnect()
         exit(1);
     }
     libvchan_close(g->vchan);
-    wait_for_possible_dispvm_resume();
     g->vchan = libvchan_server_init(g->domid, 6000, 4096, 4096);
     /* wait for gui daemon */
     while (libvchan_is_open(g->vchan) == VCHAN_WAITING)
         libvchan_wait(g->vchan);
-    send_protocol_version(g->vchan);
+    handshake(g);
     /* discard */
     read_struct(g->vchan, xconf);
     send_all_windows_info(g);
 }
 
-static void handle_sigterm()
+static _Noreturn void handle_sigterm(int UNUSED(sig),
+        siginfo_t *UNUSED(info), void *UNUSED(context))
 {
     Ghandles *g = ghandles_for_vchan_reinitialize;
     terminate_and_cleanup_xorg(g);
     exit(0);
 }
 
-static void usage()
+static void usage(void)
 {
     fprintf(stderr, "Usage: qubes_gui [options]\n");
     fprintf(stderr, "       -v  increase log verbosity\n");
@@ -2252,7 +2368,60 @@ int main(int argc, char **argv)
     int i;
     int xfd;
     Ghandles g;
-    int wait_fds[2];
+
+    g.created_input_device = access("/run/qubes-service/gui-agent-virtual-input-device", F_OK) == 0;
+
+    if(g.created_input_device) {
+        // open uinput, if it fails, falls back to xdriver to not break input to the qube
+        g.uinput_fd = open("/dev/uinput", O_WRONLY | O_NDELAY);
+        if(g.uinput_fd < 0) {
+            g.uinput_fd = open("/dev/input/uinput", O_WRONLY | O_NDELAY);
+            
+            if(g.uinput_fd < 0) {
+                fprintf(stderr, "Couldn't open uinput, falling back to xdriver\n");
+                g.created_input_device = 0;
+            }
+        }
+    }
+    
+    // input device creation
+    if(g.created_input_device) {
+        
+        if (ioctl(g.uinput_fd, UI_SET_EVBIT, EV_SYN) < 0) {
+            fprintf(stderr, "error setting EVBIT for EV_SYN, falling back to xdriver\n");
+            g.created_input_device = 0;
+        }
+
+        if (ioctl(g.uinput_fd, UI_SET_EVBIT, EV_KEY) < 0) {
+            fprintf(stderr, "error setting EVBIT for EV_KEY, falling back to xdriver\n");
+            g.created_input_device = 0;
+        }
+
+        // set all keys
+        for(int i = 1; i < KEY_MAX; i++) {
+            if((i < BTN_MISC || i > BTN_GEAR_UP) && i != KEY_RESERVED && ioctl(g.uinput_fd, UI_SET_KEYBIT, i) < 0) {
+                fprintf(stderr, "Not able to set KEYBIT %d\n", i);
+            }
+        }
+        
+        struct uinput_setup usetup;
+        memset(&usetup, 0, sizeof(usetup));
+        strcpy(usetup.name, "Qubes Virtual Input Device");
+        
+        if(ioctl(g.uinput_fd, UI_DEV_SETUP, &usetup) < 0) {
+            fprintf(stderr, "Input device setup failed, falling back to xdriver\n");
+            g.created_input_device = 0;
+        } else {
+            
+            if(ioctl(g.uinput_fd, UI_DEV_CREATE) < 0) {
+                fprintf(stderr, "Input device creation failed, falling back to xdriver\n");
+                g.created_input_device = 0;
+            }
+        }
+        
+        g.last_known_modifier_states = 0;
+    }
+
 
     parse_args(&g, argc, argv);
 
@@ -2272,7 +2441,7 @@ int main(int argc, char **argv)
         libvchan_wait(g.vchan);
     saved_argv = argv;
     vchan_register_at_eof(handle_guid_disconnect);
-    send_protocol_version(g.vchan);
+    handshake(&g);
     g.x_pid = get_xconf_and_run_x(&g);
     mkghandles(&g);
     ghandles_for_vchan_reinitialize = &g;
@@ -2312,11 +2481,17 @@ int main(int argc, char **argv)
             XFixesSelectCursorInput(g.display, RootWindow(g.display, i),
                                     XFixesDisplayCursorNotifyMask);
     } else
-        fprintf(stderr, "XFixes not available, cursor shape handling off");
+        fprintf(stderr, "XFixes not available, cursor shape handling off\n");
 
     XAutoRepeatOff(g.display);
     signal(SIGCHLD, SIG_IGN);
-    signal(SIGTERM, handle_sigterm);
+    struct sigaction sigterm_handler = {
+        .sa_sigaction = handle_sigterm,
+        .sa_flags = SA_SIGINFO,
+    };
+    sigemptyset(&sigterm_handler.sa_mask);
+    if (sigaction(SIGTERM, &sigterm_handler, NULL))
+        err(1, "sigaction");
     windows_list = list_new();
     embeder_list = list_new();
     XSetErrorHandler(dummy_handler);
@@ -2345,16 +2520,19 @@ int main(int argc, char **argv)
                     "Acquired MANAGER selection for tray\n");
     }
     xfd = ConnectionNumber(g.display);
-    wait_fds[0] = xfd;
-    wait_fds[1] = g.xserver_fd;
+    struct pollfd fds[] = {
+        { .fd = -1, .events = POLLIN | POLLHUP, .revents = 0 },
+        { .fd = xfd, .events = POLLIN | POLLHUP, .revents = 0 },
+        { .fd = g.xserver_fd, .events = POLLIN | POLLHUP, .revents = 0 },
+    };
     for (;;) {
         int busy;
-        fd_set retset;
 
-        wait_for_vchan_or_argfd(g.vchan, 2, wait_fds, &retset);
+        fds[0].fd = libvchan_fd_for_select(g.vchan);
+        wait_for_vchan_or_argfd(g.vchan, fds, QUBES_ARRAY_SIZE(fds));
         /* first process possible qubes_drv reconnection, otherwise we may be
          * using stale g.xserver_fd */
-        if (FD_ISSET(g.xserver_fd, &retset)) {
+        if (fds[2].revents) {
             char discard[64];
             int ret;
 
@@ -2370,7 +2548,7 @@ int main(int argc, char **argv)
                         "qubes_drv disconnected, waiting for possible reconnection\n");
                 close(g.xserver_fd);
                 wait_for_unix_socket(&g);
-                wait_fds[1] = g.xserver_fd;
+                fds[2].fd = g.xserver_fd;
             } else {
                 perror("reading from qubes_drv");
                 exit(1);
